<?php

/**
 * @file
 * Theme preprocessor functions.
 */
use Drupal\mcapi\Entity\Currency;
use Drupal\mcapi\Entity\CurrencyInterface;
use Drupal\Core\Render\Markup;

/**
 * Theme preprocessor for 'wallet_stats'.
 */
function template_preprocess_wallet_stats(&$vars) {
  $wallet = $vars['element']['#wallet'];
  $unused = \Drupal::config('mcapi.settings')->get('render_unused');
  foreach ($wallet->getSummaries() as $curr_id => $data) {
    if ($unused || $data['trades']) {
      $currency = Currency::load($curr_id);
      $vars['currencies'][$curr_id]['label'] = $currency->label();
      $vars['currencies'][$curr_id]['gross_in'] = $currency->format($data['gross_in'], CurrencyInterface::DISPLAY_NORMAL);
      $vars['currencies'][$curr_id]['gross_out'] = $currency->format($data['gross_out'], CurrencyInterface::DISPLAY_NORMAL);
      $vars['currencies'][$curr_id]['volume'] = $currency->format($data['volume'], CurrencyInterface::DISPLAY_NORMAL);
      $vars['currencies'][$curr_id]['balance'] = $currency->format($data['balance'], CurrencyInterface::DISPLAY_NORMAL);
      $vars['currencies'][$curr_id]['partners'] = $data['partners'];
      $vars['currencies'][$curr_id]['trades'] = $data['trades'];
    }
  }
}

/**
 * Theme preprocessor for 'wallet_wrapper'.
 */
function template_preprocess_mcapi_wallet(&$vars) {
  $vars['canonical'] = $vars['element']['#mcapi_wallet']->toUrl('canonical')->toString();
  $vars['attributes']['class'] = ['wallet', $vars['element']['#view_mode']];
}

/**
 * Theme preprocessor for 'wallet_wrapper'.
 */
function template_preprocess_mcapi_wallet_component(&$vars) {
  $vars['title'] = $vars['element']['#title'];
}

/**
 * Theme preprocessor for all of an entity's wallets.
 */
function template_preprocess_mcapi_wallets(&$vars) {
  $viewbuilder = \Drupal::entityTypeManager()->getViewBuilder('mcapi_wallet');
  // Convert each wallet into a render array.
  foreach ($vars['wallets'] as $key => $wallet) {
    $vars['wallets'][$key] = $viewbuilder->view($wallet, $vars['view_mode']);
  }
}

/**
 * Theme preprocessor for 'balances'.
 */
function template_preprocess_wallet_balances(&$vars) {
  $settings = \Drupal::config('mcapi.settings');
  $unused = $settings->get('render_unused');
  $wallet = $vars['element']['#wallet'];
  $delimited = [];
  foreach ($wallet->getSummaries() as $curr_id => $data) {
    if ($unused || $data['trades']) {
      $currency = Currency::load($curr_id);
      $vars['currencies'][$curr_id]['label'] = $currency->label();
      $balance = $currency->format($data['balance'], CurrencyInterface::DISPLAY_NORMAL, TRUE);
      // @todo even better if we can let the html through
      $balance = strip_tags($balance);
      $vars['currencies'][$curr_id]['balance'] = $balance;
      $delimited[] = $balance;
    }
  }
  $delimited = implode($settings->get('worths_delimiter'), $delimited);
  $vars['delimited'] = Markup::create($delimited);
}

/**
 * Theme preprocessor for 'balance_bars'.
 */
function template_preprocess_wallet_balance_bars(&$vars) {
  $wallet = $vars['element']['#wallet'];
  foreach ($wallet->getSummaries() as $curr_id => $data) {
    $currency = Currency::load($curr_id);
    if (!$data['gross_in'] && !$data['gross_out']) {
      return;
    }
    $id = "given-received-" . $curr_id . '-' . $wallet->id();
    $max = _mcapi_get_axis_max(max($data['gross_in'], $data['gross_out']));
    $vars['currencies'][$curr_id] = [
      'label' => $currency->label(),
      'id' => $id,
      'functionname' => str_replace('-', '_', $id),
      'incoming' => $data['gross_in'],
      'outgoing' => $data['gross_out'],
      'show_in' => $currency->format($data['gross_in'], CurrencyInterface::DISPLAY_NORMAL),
      'show_out' => $currency->format($data['gross_out'], CurrencyInterface::DISPLAY_NORMAL),
      'given' => t('Given'),
      'received' => t('Received'),
      'max' => $max,
      'vaxislabels' => [
        [
          'value' => 0,
          'label' => $currency->format(0, CurrencyInterface::DISPLAY_NORMAL),
        ],
        [
          'value' => $max / 2,
          'label' => $currency->format(intval($max / 2), CurrencyInterface::DISPLAY_NORMAL),
        ],
        [
          'value' => $max,
          'label' => $currency->format($max, CurrencyInterface::DISPLAY_NORMAL),
        ],
      ],
    ];
    foreach ($vars['currencies'][$curr_id]['vaxislabels'] as &$label) {
      $label['label'] = strip_tags($label['label']);
    }
  }
}

/**
 * Helper.
 *
 * @param int $val
 *   The max value to represent in the chart.
 */
function _mcapi_get_axis_max($val) {
  $scale = [1, 2, 5, 10, 25, 50, 100, 250, 500, 1000, 2500, 10000, 25000, 50000, 100000, 250000, 500000, 1000000, 5000000];
  $scale[] = $val;
  sort($scale);
  return $scale[array_search($val, $scale) + 1];
}

/**
 * Theme preprocessor for 'mcapi_transaction'.
 */
function template_preprocess_mcapi_transaction(&$vars, $hook, $info) {
  $element = $vars['mcapi_transaction'];
  if (isset($element['#watermark'])) {
    $vars['watermark'] = Markup::create('<div class = "watermark">'.$element['#watermark'].'</div>');
  }
  $vars['links'] = $element['#links'];
  $transaction = $element['#mcapi_transaction'];
  // We render all these fields manually because its quicker and cleaner for use in the twig file.
  $vars['payer'] = $transaction->payer->entity->tolink()->toString();
  $vars['payee'] = $transaction->payee->entity->tolink()->toString();
  $vars['description'] = Markup::create($transaction->description->value);
  $vars['state'] = $transaction->state->entity->label();
  $vars['type'] = $transaction->type->entity->label();
  $vars['creator'] = $transaction->creator->entity->toLink()->toString();

  $date_formatter = \Drupal::service('date.formatter');
  $vars['created'] = $date_formatter->format($transaction->getCreatedTime());
  $vars['changed'] = $date_formatter->format($transaction->getChangedTime());

  $renderer = \Drupal::service('renderer');
  $worth = $transaction->worth->view(['label' => 'hidden']);
  $vars['worth'] = Markup::create('<div class = "worth">' . $renderer->render($worth).'</div>');

  // find and render the fieldAPI fields
  $definitions = \Drupal::service('entity_field.manager')
    ->getFieldDefinitions('mcapi_transaction', 'mcapi_transaction');
  foreach (array_diff(array_keys($definitions), array_keys($vars), ['uuid', 'parent', 'xid']) as $fname) {
    $renderable = $transaction->{$fname}->view(['label' => 'hidden']);
    $vars[$fname] = $renderer->render($renderable);
  }

  if ($transaction->children) {
    $children = \Drupal::entityTypeManager()
      ->getViewBuilder('mcapi_transaction')
      ->viewMultiple($transaction->children, 'sentence');
    $vars['children'] = $renderer->render($children);
  }
}

/**
 * Theme preprocessor for 'mcapi_timeline'.
 *
 * Adds a column to the data with the formatted currency value.
 */
function template_preprocess_mcapi_timeline(&$vars) {
  $vars['id'] = preg_replace('/[^a-z0-9_]+/', '_', strtolower($vars['title']));
  foreach ($vars['points'] as $timestamp => $val) {
    $vars['daterows'][] = [
    // Format for javascript dates.
      date('Y,m,d', $timestamp),
      $val,
    ];
  }
  // TEMP until gcharts works better. Force some formatted values on the axis.
  if ($currency = $vars['currency']) {
    list($min, $middle, $max) = gchart_axes($vars['points']);
    $vars['vaxislabels'] = [
      [
        'value' => $min,
        'label' => $currency->format($min, CurrencyInterface::DISPLAY_NORMAL),
      ],
      [
        'value' => $middle,
        'label' => $currency->format($middle, CurrencyInterface::DISPLAY_NORMAL),
      ],
      [
        'value' => $max,
        'label' => $currency->format($max, CurrencyInterface::DISPLAY_NORMAL),
      ],
    ];
  }
  // @todo is there a way to have html labels in google charts?
  if ($vars['vaxislabels']) {
    foreach ($vars['vaxislabels'] as &$tick) {
      $tick['label'] = strip_tags($tick['label']);
    }
  }
}

/**
 * Calculate some good axis labels, using the min & max balance extents.
 *
 * @param array $vals
 *   The $points history of balances keyed by unixtimes.
 *
 * @return array
 *   Values keyed by min, 0 & max.
 */
function gchart_axes($vals) {
  $max = max($vals);
  $min = min($vals);
  if ($min >= 0) {
    $max = _mcapi_get_axis_max($max);
    return [0, $max / 2, $max];
  }
  elseif ($max <= 0) {
    $min = -_mcapi_get_axis_max(abs($min));
    return [$min, $min / 2, 0];
  }
  else {
    return [-_mcapi_get_axis_max(abs($min)), 0, _mcapi_get_axis_max($max)];
  }
}

function template_preprocess_mcapi_currency(&$variables) {
  $currency = $variables['elements']['#mcapi_currency'];
  $stats = \Drupal::entityTypeManager()
    ->getStorage('mcapi_transaction')
    ->ledgerStateQuery($currency->id(), [])
    ->execute()
    ->fetch();
  $variables['name'] = $currency->label();
  $variables['volume'] = $currency->format($stats->volume, CurrencyInterface::DISPLAY_NORMAL, TRUE);
  $variables['trades'] = $stats->trades;
  $variables['wallets'] = $stats->partners;
  $variables['url'] = $currency->toUrl();
}

