<?php

/**
 * @file
 * Token hooks.
 */
use Drupal\mcapi\Entity\CurrencyInterface;
use Drupal\Core\Render\Markup;

/**
 * Implements hook_tokens().
 */
function mcapi_tokens($type, $tokens, array $data = [], array $options = [], $bubbleable_metadata) {

  if (!array_key_exists($type, $data) || !is_object($data[$type])) {
    return;
  }
  $replacements = [];
  $token_service = \Drupal::token();
  if ($type == 'mcapi_transaction' or $type == 'xaction') {
    $transaction = $data[$type];
    foreach ($tokens as $name => $original) {
      switch ($name) {
        case 'state':
          $replacements[$original] = $transaction->state->entity->label();
          break;

        case 'type':
          $replacements[$original] = $transaction->type->entity->label();
          break;

        case PAYER_FIELDNAME:
        case PAYEE_FIELDNAME:
          $replacements[$original] = $transaction->{$name}->entity->toLink()->toString();
          break;

        case 'payee-holder':
        case 'payer-holder':
          $field = substr($name, 0, 5);
          $replacements[$original] = $transaction->{$field}->entity->getHolder()->toLink()->toString();
          break;

        case 'creator':
          $replacements[$original] = $transaction->creator->entity->toLink()->toString();
          break;

        case 'description':
          $renderable = $transaction->description->view(['label' => 'hidden']);
          $replacements[$original] = trim(strip_tags(\Drupal::service('renderer')->renderPlain($renderable)));
          break;

        case 'created':
          $replacements[$original] = format_date($transaction->created->value, 'medium');
          break;

        case 'changed':
          $replacements[$original] = format_date($transaction->changed->value, 'medium');
          break;

        case 'url':
          $replacements[$original] = $transaction->toUrl('canonical', ['absolute' => TRUE]);
          break;

        case 'worth':
          $replacements[$original] = Markup::create($transaction->worth->format(CurrencyInterface::DISPLAY_NORMAL));
          break;

        case 'sentence':
          $renderable = \Drupal::entityTypeManager()
            ->getViewBuilder('mcapi_transaction')
            ->view($transaction, 'sentence');
          $replacements[$original] = \Drupal::service('renderer')->renderPlain($renderable);
      }
    }

    // Handle any entity references - tokens within tokens.
    if ($payer_tokens = $token_service->findWithPrefix($tokens, 'payer')) {
      $replacements += $token_service->generate(
        'mcapi_wallet',
        $payer_tokens,
        ['mcapi_wallet' => $transaction->payer->entity],
        $options,
        $bubbleable_metadata
      );
    }
    if ($payee_tokens = $token_service->findWithPrefix($tokens, 'payee')) {
      $replacements += $token_service->generate(
        'mcapi_wallet',
        $payee_tokens,
        ['mcapi_wallet' => $transaction->payee->entity],
        $options,
        $bubbleable_metadata
      );
    }
    if ($creator_tokens = $token_service->findWithPrefix($tokens, 'creator')) {
      $replacements += $token_service->generate(
        'user',
        $creator_tokens,
        ['user' => $transaction->creator->entity],
        $options,
        $bubbleable_metadata
      );
    }
    if ($created_tokens = $token_service->findWithPrefix($tokens, 'created')) {
      $replacements += $token_service->generate(
        'created',
        $created_tokens,
        ['created' => $transaction->created->value],
        $options,
        $bubbleable_metadata
      );
    }
    if ($changed_tokens = $token_service->findWithPrefix($tokens, 'changed')) {
      $replacements += $token_service->generate(
        'changed',
        $changed_tokens,
        ['changed' => $transaction->changed->value],
        $options,
        $bubbleable_metadata
      );
    }
  }
  if ($type == 'wallet' or $type == 'mcapi_wallet') {
    $wallet = $data[$type];
    foreach ($tokens as $name => $original) {
      switch ($name) {
        case 'name':
          $replacements[$original] = $wallet->label();
          break;
        case 'owner':
          $replacements[$original] = $wallet->getOwner()->label();
          break;
        case 'holder':
          $replacements[$original] = $wallet->getHolder()->label();
          break;
        case 'holder-url':
          $holder = $wallet->getHolder();
          $replacements[$original] = $holder->toUrl('canonical', ['absolute' => TRUE])->toString();
          break;
        case 'url':
          $replacements[$original] = $wallet->toUrl('canonical', ['absolute' => TRUE]);
          break;
        case 'balance':
          $replacements[$original] = Markup::create($wallet->balance);
          break;
        case 'volume':
          $replacements[$original] = Markup::create($wallet->volume);
          break;
        case 'gross_in':
          $replacements[$original] = Markup::create($wallet->gross_in);
          break;
        case 'gross_out':
          $replacements[$original] = Markup::create($wallet->gross_out);
          break;
        case 'partners':
          $replacements[$original] = $wallet->partners->getValue();
          break;
        case 'trades':
          $replacements[$original] = $wallet->trades->getValue();
          break;
        case 'balances':
        case 'balance_bars':
        case 'stats':
          $renderable = [
            '#title' => t('Wallet: %name', ['%name' => $wallet->label()]),
            '#theme' => 'wallet_'.$name,
            '#wallet' => $wallet,
            '#attributes' => ['class' => ['component', $name]],
            '#theme_wrappers' => ['mcapi_wallet_component']
          ];
          $replacements[$original] = \Drupal::service('renderer')->render($renderable);
          break;
        case 'histories':
          $renderable = [
            '#title' => t('Balance history'),
            '#type' => 'balance_histories',
            '#wallet' => $wallet,
            '#attributes' => ['class' => ['component', 'histories']],
            '#width' => 300,
          ];
          $replacements[$original] = \Drupal::service('renderer')->render($renderable);
          break;

      }
    }
    if ($holder_tokens = $token_service->findWithPrefix($tokens, 'holder')) {
      $replacements += $token_service->generate(
        'holder',
        $holder_tokens,
        ['holder' => $wallet->holder()],
        $options,
        $bubbleable_metadata
      );
    }
    if ($owner_tokens = $token_service->findWithPrefix($tokens, 'owner')) {
      $replacements += $token_service->generate(
        'user',
        $owner_tokens,
        ['user' => $wallet->getOwner()],
        $options,
        $bubbleable_metadata
      );
    }
  }

  return $replacements;
}

/**
 * Implements hook_token_info().
 */
function mcapi_token_info() {
  return [
    'types' => [
      'mcapi_transaction' => [
        'name' => t('Transaction'),
        'description' => t('Tokens related to transactions'),
        'needs-data' => 'mcapi_transaction',
      ],
      'mcapi_wallet' => [
        'name' => t('Wallet'),
        'description' => t('Tokens related to wallets'),
        'needs-data' => 'mcapi_wallet',
      ]
    ],
    'tokens' => [
      'mcapi_transaction' => [
        'description' => [
          'name' => t('Description'),
        ],
        'payer' => [
          'name' => t("Payer"),
          'type' => 'wallet',
        ],
        'payee' => [
          'name' => t("payee"),
          'type' => 'wallet',
        ],
        'worth' => [
          'name' => t('Amount')
        ],
        'type' => [
          'name' => t('Workflow type'),
        ],
        'state' => [
          'name' => t('Workflow state'),
        ],
        'creator' => [
          'name' => t("Creator"),
          'type' => 'user',
        ],
        'created' => [
          'name' => t("Date created"),
          'type' => 'date',
        ],
        'changed' => [
          'name' => t("Date changed"),
          'description' => t("The date the node was most recently updated."),
          'type' => 'date',
        ],
        'sentence' => [
          'name' => t('Sentence summary')
        ],
        'url' => [
          'name' => t('Absolute url')
        ]
      ],
      'mcapi_wallet' => [
        'name' => [
          'name' => t('Name'),
        ],
        'owner' => [
          'name' => t("Owner"),
          'type' => 'user',
        ],
        'url' => [
          'name' => t('Absolute url'),
        ],
        'holder' => [
          'name' => t('Holder label'),
        ],
        'balance' => [
          'name' => t('Balance'),
        ],
        'volume' => [
          'name' => t('Volume'),
        ],
        'gross_in' => [
          'name' => t('Gross income'),
        ],
        'gross_out' => [
          'name' => t('Gross expenditure'),
        ],
        'trades' => [
          'name' => t('Trades'),
        ],
        'partners' => [
          'name' => t('Partners'),
        ],
        'holder-url' => [
          'name' => t('Absolute url of holder'),
        ],
        'balances' => [
          'name' => t('All balances in a table'),
        ],
        'balance_bars' => [
          'name' => t('Income & expenditure'),
        ],
        'stats' => [
          'name' => t('All stats in one table'),
        ],
        'histories' => [
          'name' => t('Balance history'),
        ]
      ]
    ]
  ];
  return $info;
}

function mcapi_token_info_alter(&$info) {
  // Make aliases for transactions & wallets
  $info['tokens']['xaction'] = $info['tokens']['mcapi_transaction'];
  $info['types']['xaction'] = $info['types']['mcapi_transaction'];
  $info['tokens']['wallet'] = $info['tokens']['mcapi_wallet'];
  $info['types']['wallet'] = $info['types']['mcapi_wallet'];
}