<?php

namespace Drupal\mcapi_forms\Plugin\Field\FieldWidget;

use Drupal\mcapi\Entity\Wallet;
use Drupal\user\Entity\User;
use \Drupal\Core\Session\AccountInterface;
use Drupal\Core\Messenger\MessengerInterface;
use Drupal\Core\Entity\ContentEntityStorageInterface;
use Drupal\Core\Field\Plugin\Field\FieldWidget\OptionsButtonsWidget;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Field\FieldItemListInterface;
use Symfony\Component\Validator\ConstraintViolationInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Plugin implementation of the wallet selector widget.
 *
 * @FieldWidget(
 *   id = "my_wallet",
 *   label = @Translation("My wallet"),
 *   description = @Translation("Choose from the current user's wallets"),
 *   field_types = {
 *     "wallet_reference"
 *   }
 * )
 * @todo inject current user, $messenger
 */
class MyWalletWidget extends OptionsButtonsWidget {

  protected $currentUser;
  protected $walletStorage;
  protected $messenger;

  /**
   *
   * @param type $plugin_id
   * @param type $plugin_definition
   * @param type $field_definition
   * @param type $settings
   * @param type $third_party_settings
   * @param AccountInterface $current_user
   * @param ContentEntityStorageInterface $wallet_storage
   * @param MessengerInterface $messenger
   */
  public function __construct($plugin_id, $plugin_definition, $field_definition, $settings, $third_party_settings, AccountInterface $current_user, ContentEntityStorageInterface $wallet_storage, MessengerInterface $messenger) {
    parent::__construct($plugin_id, $plugin_definition, $field_definition, $settings, $third_party_settings);
    $this->currentUser = $current_user;
    $this->walletStorage = $wallet_storage;
    $this->messenger = $messenger;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $plugin_id,
      $plugin_definition,
      $configuration['field_definition'],
      $configuration['settings'],
      $configuration['third_party_settings']?? [],
      $container->get('current_user'),
      $container->get('entity_type.manager')->getStorage('mcapi_wallet'),
      $container->get('messenger')
    );
  }

  /**
   * {@inheritdoc}
   */
  public static function defaultSettings() {
    return parent::defaultSettings() + [
      'hide_one_wallet' => TRUE,
    ];
  }

  /**
   * {@inheritdoc}
   */
  public function settingsForm(array $form, FormStateInterface $form_state) {
    $element = parent::settingsForm($form, $form_state);
    unset($element['size']);
    $element['hide_one_wallet'] = [
      '#title' => $this->t('Hide this field if there is only one wallet available.'),
      '#type' => 'checkbox',
      '#default_value' => $this->getSetting('hide_one_wallet'),
    ];
    return $element;
  }

  /**
   * {@inheritdoc}
   */
  public function settingsSummary() {
    $summary = parent::settingsSummary();
    // Size.
    unset($summary[1]);
    $message = $this->getSetting('hide_one_wallet') ?
      $this->t('Hide widget for one wallet') :
      $this->t('Show widget for one wallet');
    $summary['hide_one'] = ['#markup' => $message];
    return $summary;
  }

  /**
   * {@inheritdoc}
   */
  public function formElement(FieldItemListInterface $items, $delta, array $element, array &$form, FormStateInterface $form_state) {
    // My wallets are the ones held by me.
    $current_user = User::load($this->currentUser->id());
    $wallet_ids = $this->walletStorage->walletsOf($current_user);
    $count = count($wallet_ids);
    if (!$count) {
      $this->messenger->addError($this->t('No wallets to show for @role', ['@role' => $this->fieldDefinition->getLabel()]));
      $form['#disabled'] = TRUE;
      return [];
    }
    if ($count == 1) {
      $wid = reset($wallet_ids);
      $element['#value'] = $wid;
      if ($this->getSetting('hide_one_wallet')) {
        $element['#type'] = 'value';
      }
      else {
        $element['#type'] = 'item';
        $element['#markup'] = Wallet::load($wid)->label();
      }
    }
    else {
      $element['#type'] = 'radios';
      foreach ($wallet_ids as $wid) {
        $element['#options'][$wid] = Wallet::load($wid)->label();
        $element['#pre_render'] = [];
      }
    }
    return ['target_id' => $element];
  }

  /**
   * {@inheritdoc}
   */
  public function errorElement(array $element, ConstraintViolationInterface $error, array $form, FormStateInterface $form_state) {
    return isset($element['target_id']) ? $element['target_id'] : FALSE;
  }

}
