<?php

namespace Drupal\mcapi_signatures\Form;

use Drupal\Core\Form\ConfigFormBase;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\mcapi\Entity\Type;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Form for signature settings.
 */
class Settings extends ConfigFormBase {

  private $transactionRelativeManager;

  /**
   * Constructs a \Drupal\system\ConfigFormBase object.
   *
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   *   The factory for configuration objects.
   */
  public function __construct(ConfigFactoryInterface $config_factory, $transactionRelativeManager) {
    parent::__construct($config_factory);
    $this->transactionRelativeManager = $transactionRelativeManager;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('config.factory'),
      $container->get('mcapi.transaction_relative_manager')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'mcapi_signatures_settings_form';
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {
    // @todo This would look really tidy in a grid - but forms in tables are tricky
    $options = $this->transactionRelativeManager->options();
    unset($options['signatory'], $options['pending_signatory']);
    $i= 0;
    $form['intro']['#markup'] = t("For each transaction type, specify which of the transactions relatives must to sign it before the transaction moves to workflow state 'completed'") . ' '.
      t("Each mail template MUST include the token [xaction:sign-link]");
    foreach (Type::loadMultiple() as $type) {
      $id = $type->id();
      $relations = $type->getThirdPartySetting('mcapi_signatures', 'relations', []);
      $mail = $type->getThirdPartySetting('mcapi_signatures', 'mail', ['subject' => '', 'body' => '']);
      $form[$id] = [
        '#title' => $this->t("Settings for '@type' transactions", ['@type' => $type->label]),
        '#description' => $type->description,
        '#type' => 'details',
        '#collapsible' => TRUE,
        '#tree' => TRUE,
        '#open' => (bool)$relations,
        'relations' => [
          '#title' => t('Signatories'),
          '#type' => 'checkboxes',
          '#options' => $options,
          '#default_value' => $relations,
          '#required' => FALSE,
          '#weight' => 1,
        ],
        'mail' => [
          'subject' => [
            '#title' => t('Mail subject'),
            '#type' => 'textfield',
            '#default_value' => $mail['subject'],
            '#weight' => 1,
          ],
          'body' => [
            '#title' => t('Mail body'),
            '#type' => 'textarea',
            '#weight' => $i++,
            '#default_value' => $mail['body'],
            '#weight' => 1,
          ],
          '#weight' => 2,
        ]
      ];
    }
    if (\Drupal::moduleHandler()->moduleExists('token')) {
      $form['tokens'] = \Drupal::service('token.tree_builder')
        ->buildRenderable(['user', 'xaction', 'site', 'current-date'], ['global_types' => FALSE]);
    }
    return parent::buildForm($form, $form_state);
  }

  /**
   * {@inheritdoc}
   */
  public function validateForm(array &$form, FormStateInterface $form_state) {
    // This is required.
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    $form_state->cleanValues();
    foreach ($form_state->getValues() as $type_name => $vals) {
      if ($type_name == 'tokens') {
        continue;
      }
      $type = Type::load($type_name);
      $type->setThirdPartySetting('mcapi_signatures', 'relations', array_filter($vals['relations']));
      $type->setThirdPartySetting('mcapi_signatures', 'mail', $vals['mail']);
      $type->save();
    }
    parent::submitForm($form, $form_state);
    $form_state->setRedirect('mcapi.admin.workflow');
  }

  /**
   * {@inheritdoc}
   */
  public function getEditableConfigNames() {}

}
