<?php

namespace Drupal\mcapi\Entity\Storage;

use Drupal\Core\Entity\EntityInterface;
use Drupal\mcapi\Entity\Transaction;

/**
 * Storage controller for Transaction entity.
 */
class TransactionStorage extends TransactionIndexStorage {

  /**
   * {@inheritdoc}
   */
  public function loadBySerial($serial, $exception_fail = FALSE) {
    $transactions = $this->loadByProperties(['serial' => $serial]);
    ksort($transactions);
    if ($transactions) {
      $transaction = array_shift($transactions);
      $transaction->children = $transactions;
      return $transaction;
    }
    if ($exception_fail) {
      throw new \Exception('No transaction with serial '.$serial);
    }
  }


  /**
   * {@inheritdoc}
   *
   * Because the transaction entity is keyed by serial number not xid,
   * and because it contains child entities,
   * We need to overwrite the whole save function
   * and by the time we call the parent, we pass it individual transaction
   * entities having called $transaction->flatten.
   */
  public function save(EntityInterface $transaction) {
    // Determine the serial number.
    if (empty($transaction->serial->value)) {
      $serial = $this->database->query(
        "SELECT MAX(serial) FROM {mcapi_transaction}"
      )->fetchField() + 1;
    }
    else {
      $serial = $transaction->serial->value;
    }
    // Note that flatten() clones the transactions.
    foreach ($transaction->flatten(FALSE) as $entity) {
      $entity->serial->value = $serial;
      // Entity parent is 0 for the first entity and then the xid of the first
      // for all children saved consequently.
      $entity->parent->target_id = $parent??0;
      $return = parent::save($entity);
      if (!isset($parent)) {
        $parent = $entity->id();
      }
    }

    $transaction->serial->value = $serial;
    // Allow code to run after saving.
    $transaction->setOriginalId($transaction->id());
    return $return;
  }

  /**
   * {@inheritdoc}
   */
  public function doSave($id, EntityInterface $entity) {
    $record = $this->mapToStorageRecord($entity);
    $record->changed = REQUEST_TIME;
    $return = parent::doSave($entity->xid->value, $entity);
    // The entity is no longer new.
    // because we were working on a clone.
    $entity->enforceIsNew(FALSE);

    return $return;
  }

  /**
   * {@inheritdoc}
   */
  public function resetCache(array $ids = NULL) {
    parent::resetCache($ids);
  }

  /**
   * {@inheritdoc}
   */
  protected function doDelete($entities) {
    foreach ($entities as $transaction) {
      $this->invalidateWalletCache($transaction);
    }
    parent::doDelete($entities);
  }

  /**
   * @param EntityInterface $entity
   * @param bool $update
   */
  protected function doPostSave(EntityInterface $entity, $update) {
    $this->invalidateWalletCache($entity);
    parent::doPostSave($entity, $update);
  }

  /**
   * @param Transaction $transaction
   */
  private function invalidateWalletCache(Transaction $transaction) {
    foreach ($transaction->flatten() as $t) {
      // in cases of migration rollbacks the wallet may not be present
      if ($payer_wallet = $t->payer->entity) {
        $payer_wallet->invalidateTagsOnSave(TRUE);
      }
      if ($payee_wallet = $t->payee->entity) {
        $payee_wallet->invalidateTagsOnSave(TRUE);
      }
    }
  }

}
