<?php

namespace Drupal\mcapi\EventSubscriber;

use Drupal\mcapi\Entity\Currency;
use Drupal\mcapi\Entity\Wallet;
use Drupal\user\Entity\Role;
use Drupal\migrate\Plugin\Migration;
use Drupal\migrate\Event\MigrateImportEvent;
use Drupal\migrate\Event\MigratePreRowSaveEvent;
use Drupal\migrate\Event\MigrateRollbackEvent;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;

/**
 * Alterations to migrations
 */
class MigrationSubscriber implements EventSubscriberInterface {

  /**
   * {@inheritdoc}
   */
  public static function getSubscribedEvents() {
    return [
      'migrate.pre_import' => [['migratePreImport']],
      'migrate.post_import' => [['migratePostImport']],
      'migrate.pre_row_save' => [['migratePreRowSave']],
      'migrate.post_rollback' => [['migratePostRollback']]
    ];
  }

  /**
   * Subscribed event callback.
   *
   * @param MigratePreRowSaveEvent $event
   *
   * Change the name of the transaction description field, which was a vairable
   * in d7, but an entity basefield in d8. Change the names of the transaction
   * entitytype and bundle.
   */
  public function migratePreRowSave(MigratePreRowSaveEvent $event) {
    $row = $event->getRow();
    $migration = $event->getMigration();

    // Change the name of the transaction entity and bundle.
    if ($migration->id() == 'd7_field' or $migration->id() == 'd7_field_instance') {
      if ($row->getDestinationProperty('entity_type') == 'transaction') {
        $row->setDestinationProperty('entity_type', 'mcapi_transaction');
        if ($row->getSourceProperty('plugin') == 'd7_field_instance') {
          $row->setDestinationProperty('bundle', 'mcapi_transaction');
        }
      }
    }

    // Rename the transaction description field
    if ($migration->id() == 'd7_mcapi_transaction') {
      $old_desc_field_name = $this->d7DescriptionFieldname($migration);
      // Rename the transaction description field
      $row->setDestinationProperty(
        'description',
        $row->getSourceProperty($old_desc_field_name)
      );
      $row->removeDestinationProperty($old_desc_field_name);
    }
    // Horrible but necessary if we change the entity name
    if ($migration->getDestinationConfiguration()['plugin'] == 'component_entity_display') {
      if ($row->getDestinationProperty('entity_type') == 'transaction') {
        $row->setDestinationProperty('entity_type', 'mcapi_transaction');
        $row->setDestinationProperty('bundle', 'mcapi_transaction');
      }
    }

    if ($event->getMigration()->id() == 'd7_mcapi_form') {
      if ($row->getSourceProperty('name') == '1stparty') {
        $direction = $row->getSourceProperty('direction');
        if ($direction->preset == 'outgoing') {
          $row->setDestinationProperty('mode',  'credit');
        }
        elseif ($direction->preset == 'incoming') {
          $row->setDestinationProperty('mode',  'bill');
        }
      }
    }
  }

  /**
   * Subscribed event callback.
   */
  function migratePreImport(MigrateImportEvent $event) {
    // Delete existing currencies before importing from the old site.
    if ($event->getMigration()->id() == 'd7_mcapi_currency') {
      foreach (Currency::loadMultiple() as $currency) {
        if ($currency->access('delete')) {
          $currency->delete();
        }
      }
    }
    if ($event->getMigration()->id() == 'd7_user') {
      // Delete wallets to ensure no clash with incoming wallets
      foreach (Wallet::loadMultiple() as $wallet) {
        $wallet->delete();
      }
      // Disable auto adding wallets
      $wallet_settings = \Drupal::configFactory()->getEditable('mcapi.settings');
      $autoadd = $wallet_settings->get('autoadd');
      $wallet_settings->set('autoadd', 0)->save();
      \Drupal::state()->set('user_wallet_autocreate', $autoadd);
    }
  }

  /**
   * Subscribed event callback.
   * Set the autoadd back to what it was before the user migration
   */
  function migratePostImport(MigrateImportEvent $event) {
    if ($event->getMigration()->id() == 'd7_user') {
      $autoadd = \Drupal::state()->get('user_wallet_autocreate');
      $wallet_settings = \Drupal::configFactory()
        ->getEditable('mcapi.settings')
        ->set('autoadd', $autoadd)
        ->save();
      \Drupal::state()->delete('user_wallet_autocreate');
    }
    // Finish the permission migration, because it didn't map 1:1
    if ($event->getMigration()->id() == 'd7_user_roles') {
      foreach (Role::loadMultiple() as $role) {
        if ($role->hasPermission('use wallets')) {
          $role->grantPermission('view all wallets');
          $role->grantPermission('view all transactions');
          $role->grantPermission('create remote transactions');
        }
        if ($role->hasPermission('manage mcapi')) {
          $role->grantPermission('create 3rdparty transaction');
          $role->grantPermission('create wallets');
        }
        $role->save();
      }
    }
  }

  /**
   * Subscribed event callback.
   * Set the autoadd back to what it was before the user migration
   */
  function migratePostRollback(MigrateRollbackEvent $event) {
    if ($event->getMigration()->id() == 'd7_user') {
      foreach (Wallet::loadMultiple() as $w){
        $w->delete();
      }
    }
  }

  /**
   * Utility.
   *
   * @param Migration $migration
   *
   * @return string
   *   The fieldname of the d7 transaction description
   */
  static function d7DescriptionFieldname(Migration $migration) {
    $result =  $migration->getSourcePlugin()
     ->getDatabase()
     ->select('variable', 'v')
     ->fields('v', ['value'])
     ->condition('name', 'transaction_description_field')
     ->execute()
     ->fetchField();
   return unserialize($result);
  }

}
