<?php

namespace Drupal\mcapi\Form;

use Drupal\mcapi\Mcapi;
use Drupal\user\Entity\User;
use Drupal\Core\Entity\ContentEntityForm;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Cache\Cache;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Build form to edit, not create, the wallet entity.
 */
class WalletForm extends ContentEntityForm {


  /**
   * The wallet settings.
   */
  protected $walletConfig;

  /**
   * @param EntityRepositoryInterface $entity_repository
   * @param EntityTypeBundleInfoInterface $entity_type_bundle_info
   * @param TimeInterface $time
   * @param ConfigFactoryInterface $config_factory
   */
  public function __construct($entity_repository, $entity_type_bundle_info, $time, ConfigFactoryInterface $config_factory) {
    parent::__construct($entity_repository, $entity_type_bundle_info, $time);
    $this->walletConfig = $config_factory->get('mcapi.settings');
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('entity.repository'),
      $container->get('entity_type.bundle.info'),
      $container->get('datetime.time'),
      $container->get('config.factory')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'wallet_form';
  }

  /**
   * {@inheritdoc}
   */
  public function form(array $form, FormStateInterface $form_state) {
    $wallet = $this->entity;
    $holder = $wallet->getHolder();
    $form['#title'] = $this->t("Edit %name wallet", ['%name' => $wallet->label()]);

    $form = parent::form($form, $form_state);

    $form['name'] = [
      '#title' => $this->t('Name or purpose of wallet'),
      '#description' => $this->t('The wallet name is used in the transaction form.'),
      '#type' => 'textfield',
      '#default_value' => $wallet->name->value,
      '#placeholder' => $this->t("@name - personal", ['@name' => $holder->label()]),
      '#required' => TRUE,
      '#maxlength' => 48,
      '#size' => 48,
    ];

    $walletableBundles = Mcapi::walletableBundles();
    $entity_types = [];
    foreach (array_keys($walletableBundles) as $entity_type_id) {
      $entity_types[$entity_type_id] = $this->entityTypeManager->getDefinition($entity_type_id);
    }

    $form['transfer'] = [
      '#title' => t('Change holder from %name', ['%name' => $wallet->holder->entity->label()]),
      '#type' => 'details',
      'holder_entity_type' => [
        '#title' => t('New holder entity type'),
        '#type' => 'select',
        '#options' => array_map(function ($type) {return $type->getLabel();}, $entity_types),
        '#default_value' => $wallet->holder_entity_type->value,
        '#required' => FALSE,
        '#access' => $this->currentUser()->hasPermission('manage exchange') && count($entity_types) > 1,
      ],
    ];
    foreach ($walletableBundles as $entity_type_id => $bundles) {
      $form['transfer'][$entity_type_id . '_entity_id'] = [
        '#title' => t('New holder name'),
        '#type' => 'entity_autocomplete',
        '#placeholder' => t('@entityname name...', ['@entityname' => $entity_types[$entity_type_id]->getLabel()]),
        '#default_value' => '',
        '#states' => [
          'visible' => [
            ':input[name="holder_entity_type"]' => ['value' => $entity_type_id],
          ],
        ],
        '#autocomplete_route_name' => 'system.entity_autocomplete',
        '#target_type' => $entity_type_id,
        '#selection_handler' => 'default',
      ];
      if ($entity_types[$entity_type_id]->hasKey('bundle')) {
        $form['transfer'][$entity_type_id . '_entity_id']['#selection_settings']['target_bundles'] = array_keys($bundles);
      }
    }
    $form['#attached']['library'][] = 'mcapi/mcapi.wallets';
    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function save(array $form, FormStateInterface $form_state) {
    $wallet = $this->entity;

    $values = $form_state->getValues();
    // Update the wallet with any change of access
    // check for the change of holdership.
    $key = $values['holder_entity_type'] . '_entity_id';
    if ($values[$key]) {
      // Invalidate cache tags on the current parent.
      // @see Wallet::invalidateTagsOnSave to invalidate tags on the new parent.
      Cache::invalidateTags([$wallet->holder_entity_type->value . ':' . $wallet->holder_entity_id->value]);

      $wallet->set('holder_entity_type', $values['holder_entity_type'])
        ->set('holder_entity_id', $values[$key]);
      // @todo need to clear the walletaccess cache for both users
      $this->messenger()->addStatus(
        t('Wallet has been transferred to %name', ['%name' => $wallet->getOwner()->label()])
      );
    }
    $form_state->setRedirect(
      'entity.mcapi_wallet.canonical',
      ['mcapi_wallet' => $wallet->id()]
    );

    parent::save($form, $form_state);
  }

  /**
   * Get a list of user names from the user IDs.
   *
   * @param array $uids
   *   User IDs.
   *
   * @return string
   *   Comma separated usernames.
   */
  private function getUsernames($uids) {
    $names = [];
    foreach (User::loadMultiple($uids) as $account) {
      $names[] = $account->getDisplayName();
    }
    return implode(', ', $names);
  }

}
