<?php

namespace Drupal\mcapi\Form;

use Drupal\Core\Entity\Entity\EntityFormDisplay;
use Drupal\Core\Form\ConfigFormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Cache\Cache;
use Drupal\Core\Entity\EntityType;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Form builder for wallet settings.
 */
class WalletSettings extends ConfigFormBase {

  const UNWALLETED = 0;
  const WALLETABLE = 1;
  const AUTOCREATE = 2;

  /**
   * The entityTypeManager service.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The entity type bundle info service.
   *
   * @var type
   */
  protected $entityTypeBundleInfo;

  /**
   * Construct.
   */
  public function __construct($configFactory, $entityTypeManager, $entity_type_bundle_info) {
    $this->setConfigFactory($configFactory);
    $this->entityTypeManager = $entityTypeManager;
    $this->entityTypeBundleInfo = $entity_type_bundle_info;
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId() {
    return 'mcapi_wallet_settings_form';
  }

  /**
   * {@inheritdoc}
   */
  static public function create(ContainerInterface $container) {
    return new static(
      $container->get('config.factory'),
      $container->get('entity_type.manager'),
      $container->get('entity_type.bundle.info')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) {
    // Display a warning message.
    $display = EntityFormDisplay::load('mcapi_wallet.mcapi_wallet.default');
    $config = $this->configFactory()->get('mcapi.settings');

    $form['entity_types'] = [
      '#title' => $this->t('Wallet ownership'),
      '#type' => 'fieldset',
      '#weight' => 2,
      '#tree' => TRUE,
//      'user:user' => [
//        '#title' => t('Autocreate for permitted users'),
//        '#description' => t('When a user loses permission the wallet reverts to user 1'),
//        '#type' => 'checkbox',
//        // We will add 1 to the result so the meaning corresponds with this form's constants
//        '#default_value' => (int)$config->get("entity_types.user:user") -1,
//      ]
    ];

    foreach ($this->entityTypeManager->getDefinitions() as $entity_type_id => $entity_type) {
      if ($entity_type_id == 'mcapi_transaction' or $entity_type_id == 'comment') {
        continue;
      }
      $reflection = new \ReflectionClass($entity_type->getClass());
      if ($entity_type_id == 'user' or ($reflection->isSubClassOf('Drupal\user\EntityOwnerInterface') and $entity_type->getLinkTemplate('canonical'))) {
        $bundles = $this->entityTypeBundleInfo->getBundleInfo($entity_type_id);
        $entity_label = (count($bundles) > 1)
          ? $entity_type->getLabel() . ': '
          : '';
        // Todo set the min to the most number of existing wallets per bundle.
        foreach ($bundles as $bundle_name => $bundle_info) {
          $val = (int)$config->get("entity_types.$entity_type_id:$bundle_name");
          /* @var $bundle_name string */
          $form['entity_types']["$entity_type_id:$bundle_name"] = [
            '#title' => $entity_label . $bundle_info['label'],
            '#type' => 'select',
            '#options' => $this->walletTypeOptions($val, $this->bundleHasWallets($entity_type, $bundle_name)),
            '#default_value' => $val,
          ];
        }
      }
    }

    $form['render_unused'] = [
      '#title' => $this->t('Show unused currencies'),
      '#description' => $this->t('Render stats for unused currencies.'),
      '#type' => 'checkbox',
      '#default_value' => $config->get('render_unused'),
      '#weight' => 7,
    ];
    $form['public'] = [
      '#title' => $this->t('Default wallet visibility'),
      '#description' => $this->t('Setting given to new wallets.') . ' NOT IMPLEMENTED',
      '#type' => 'radios',
      '#options' => [
        0 => $this->t('Can only be seen by the owners and admins'),
        1 => $this->t("Can be seen by anybody with 'view public wallets' permission"),
      ],
      //'#default_value' => $config->get('public'),
      '#disabled' => TRUE,
      '#weight' => 9,
    ];

    $form['user_interface'] = [
      '#title' => $this->t('User interface'),
      '#type' => 'details',
      '#weight' => 12,
    ];

    $form['user_interface']['wallet_tab'] = [
      '#title' => $this->t('Show wallets tab on canonical entity page'),
      '#description' => $this->t("Tab would show 'summary' view of each wallet owned by the entity. Otherwise show links to wallets canonical pages using the entity's display fields"),
      '#type' => 'checkbox',
      '#default_value' => $config->get('wallet_tab'),
      '#weight' => 3,
    ];
    $form['user_interface']['wallet_inex_tab'] = [
      '#title' => $this->t('Show income & expenditure tab on wallet page'),
      '#description' => $this->t("Alternatively include it in a wallet view mode"),
      '#type' => 'checkbox',
      '#default_value' => $config->get('wallet_inex_tab'),
      '#weight' => 3,
    ];
    $form['user_interface']['wallet_log_tab'] = [
      '#title' => $this->t('Show transaction log tab on wallet page'),
      '#description' => $this->t("Alternatively include it in a wallet view mode"),
      '#type' => 'checkbox',
      '#default_value' => $config->get('wallet_log_tab'),
      '#weight' => 3,
    ];

    return parent::buildForm($form, $form_state);
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state) {
    $entity_types = $form_state->getValue('entity_types');
    //$entity_types['user:user']++; // Convert from the checkbox
    $this->configFactory->getEditable('mcapi.settings')
      ->set('entity_types', array_filter($entity_types))
      ->set('wallet_tab', $form_state->getValue('wallet_tab'))
      ->set('wallet_inex_tab', $form_state->getValue('wallet_inex_tab'))
      ->set('wallet_log_tab', $form_state->getValue('wallet_log_tab'))
      ->save();

    parent::submitForm($form, $form_state);

    Cache::invalidateTags(
      ['mcapi_wallet_values', 'mcapi_wallet_view', 'walletable_bundles']
    );

    $form_state->setRedirect('mcapi.admin');
  }

  /**
   * {@inheritdoc}
   */
  protected function getEditableConfigNames() {
    return ['mcapi.settings'];
  }

  /**
   *
   * @param EntityType $entity_type
   * @param string $bundle_name
   * @param int $current_val
   * @return array
   */
  private function walletTypeOptions(int $current_val, bool $unwalleted ) : array {
    $options = [];
    if ($current_val == SELF::UNWALLETED or $unwalleted) {
      $options[SELF::UNWALLETED] = t('Never holds wallets');
    }
    $options[SELF::WALLETABLE] = t('Wallets can be given');
    $options[SELF::AUTOCREATE] = t('Created with a wallet');
    return $options;
  }

  /**
   * Find out if any entities of the given bundle already have wallets.
   * @param EntityType $entity_type
   * @param string $bundle_name
   * @return bool
   *   TRUE if the bundle has wallets
   */
  private function bundleHasWallets (EntityType $entity_type, string $bundle_name) : bool {
    $wids = $this->entityTypeManager->getStorage('mcapi_wallet')
      ->getQuery()
      ->condition('holder_entity_type', $entity_type->id())
      ->execute();
    if ($entity_type->hasKey('bundle')) {
      $bundle_wids = $this->entityTypeManager->getStorage($entity_type->id())
        ->getQuery()
        ->condition($entity_type->getKey('bundle'), $bundle_name)
        ->execute();
      $wids = array_intersect($wids, $bundle_wids);
    }
    return empty($wids);
  }
}
