<?php

namespace Drupal\mcapi;

use Drupal\mcapi\Form\WalletSettings;
use Drupal\mcapi\Plugin\TransactionRelativeManager;
use Drupal\user\EntityOwnerInterface;
use Drupal\Core\Url;
use Drupal\Core\Link;
use Drupal\Core\Entity\ContentEntityInterface;
use Drupal\Component\Utility\Html;

/**
 * Utility class for community accounting.
 */
class Mcapi {

  /**
   * Get a list of bundles which can have wallets.
   *
   * Which is to say, whether it is a contentEntity configured to have at least
   * 1 wallet.
   *
   * @return array
   *   Bundle names keyed by entity type ids.
   */
  public static function walletableBundles($reset = FALSE) :array {
    $bundles = &drupal_static(__FUNCTION__);
    if (!is_array($bundles) || $reset) {
      $bundles = [];
      foreach (\Drupal::Config('mcapi.settings')->get('entity_types') as $entity_bundle => $mode) {
        list($type, $bundle) = explode(':', $entity_bundle);
        $bundles[$type][$bundle] = $mode;
      }
    }
    return $bundles;
  }

  /**
   * @param ContentEntityInterface $holder
   * @return @bool
   */
  static function holderIsWalletable(ContentEntityInterface $holder) : bool {
    $result = FALSE;
    if ($holder->getEntityTypeId() == 'user') {
      $result = TRUE;
    }
    elseif ($holder instanceOf EntityOwnerInterface) {
      $bundles = static::walletableBundles();
      $result = isset($bundles[$holder->getEntitytypeId()][$holder->bundle()]);
    }
    return $result;
  }

  /**
   * Uasort callback for configuration entities.
   *
   * Should have been included in Drupal Core?
   */
  public static function uasortWeight($a, $b) :int {
    $a_weight = (is_object($a) && property_exists($a, 'weight')) ? $a->weight : 0;
    $b_weight = (is_object($b) && property_exists($b, 'weight')) ? $b->weight : 0;
    if ($a_weight == $b_weight) {
      return 0;
    }
    return ($a_weight < $b_weight) ? -1 : 1;
  }

  /**
   * @param EntityInterface[] $entities
   * @return array
   */
  static function entityOptionsList(array $entities) :array {
    $list = [];
    $contextRepo = \Drupal::service('entity.repository');
    foreach ($entities as $entity) {
      $label = $contextRepo->getTranslationFromContext($entity)->label();
      $list[$entity->id()] = Html::escape($label);
    }
    return $list;
  }

  /**
   * Service container.
   *
   * @return \Drupal\mcapi\Plugin\TransactionRelativeManager
   *   The 'Transaction relative manager' service.
   */
  public static function transactionRelatives($plugin_names = []) :TransactionRelativeManager {
    return \Drupal::service('mcapi.transaction_relative_manager')->activatePlugins($plugin_names);
  }

  /**
   * Get a list of all the tokens on the transaction entity.
   *
   * @note This really needs replacing with a core function.
   */
  public static function tokenHelp() :string {
    foreach (array_keys(\Drupal::Token()->getInfo()['tokens']['xaction']) as $token) {
      $tokens[] = "[xaction:$token]";
    }
    return implode(', ', $tokens);
  }

  /**
   * Get the renderable transaction fields for use in tokens or twig.
   *
   * @return []
   *   Keyed by fieldname.
   *
   * @note There are no extraFields at the moment.
   */
  public static function transactionFields() :array {
  $definitions = \Drupal::service('entity_field.manager')
    ->getFieldDefinitions('mcapi_transaction', 'mcapi_transaction');
    // Chance to remove some tokens here using unset($definitions['uuid']).
    unset($definitions['parent'], $definitions['uuid'], $definitions['xid']);
    return $definitions;
  }

  /**
   * Get a string describing where to get help writing Twig.
   */
  public static function twigHelp() :string {
    foreach (Mcapi::transactionFields() as $fieldname => $info) {
      $tokens[] = '{{ ' . $fieldname . ' }}';
    }
    // @todo how to place links in $element['#description']?
    $link = Link::fromTextAndUrl(
        t('What is @Twig?', ['@Twig' => 'Twig']),
        Url::fromUri('http://twig.sensiolabs.org/doc/templates.html')
      );
    return implode(', ', $tokens) . '. ' . $link->toString();
  }
}
