<?php

namespace Drupal\mcapi\Plugin\EntityReferenceSelection;

use Drupal\Component\Utility\Html;
use Drupal\Core\Entity\Plugin\EntityReferenceSelection\DefaultSelection;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Drupal\Core\Database\Driver\mysql\Connection;

/**
 * Provide default Wallet selection handler.
 *
 * The settings consist of one value, restriction, which can be in 3 modes
 *
 * @EntityReferenceSelection(
 *   id = "default:mcapi_wallet",
 *   label = @Translation("Wallet selection"),
 *   entity_types = {"mcapi_wallet"},
 *   group = "default"
 * )
 * @todo it looks like the restrictions are deprecated or not yet coded.
 */
class WalletSelection extends DefaultSelection {

  private $database;

  /**
   * Possible values for the restriction setting.
   */
  const RESTRICTION_PAYIN = 'payin';
  const RESTRICTION_PAYOUT = 'payout';
  const RESTRICTION_MINE = 'currentuser';
  const RESTRICTION_NONE = 'none';


  public function __construct(array $configuration, $plugin_id, $plugin_definition, $entity_type_manager, $module_handler, $current_user, $entity_field_manager = NULL, $entity_type_bundle_info = NULL, $entity_repository = NULL, Connection $database) {
    parent::__construct($configuration, $plugin_id, $plugin_definition, $entity_type_manager, $module_handler, $current_user, $entity_field_manager, $entity_type_bundle_info, $entity_repository);
    $this->database = $database;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('entity_type.manager'),
      $container->get('module_handler'),
      $container->get('current_user'),
      $container->get('entity_field.manager'),
      $container->get('entity_type.bundle.info'),
      $container->get('entity.repository'),
      $container->get('database')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function getReferenceableEntities($match = NULL, $match_operator = 'CONTAINS', $limit = 0) {
    $query = $this->walletsQuery($match);
    if ($limit > 0) {
      $query->range(0, $limit);
    }
    //Quicker than calling Wallet::loadMultiple, esp since entityTypeManager is loaded
    $entities = $this->entityTypeManager
      ->getStorage('mcapi_wallet')
      ->loadMultiple($query->execute());

    $options = ['mcapi_wallet' => []];
    foreach ($entities as $entity_id => $entity) {
      $options['mcapi_wallet'][$entity_id] = Html::escape($entity->label());
    }
    return $options;
  }

  /**
   * {@inheritdoc}
   */
  public function countReferenceableEntities($match = NULL, $match_operator = 'CONTAINS') {
    return $this->walletsQuery()->count()->execute();
  }

  /**
   * {@inheritdoc}
   */
  public function validateReferenceableEntities(array $ids) {
    $valid_ids = $this->walletsQuery()->execute();
    // User 1 skips validation. this is helpful for importing.
    if ($this->currentUser->id() != 1) {
      $ids = array_intersect($ids, $valid_ids);
    }
    return $ids;
  }

  /**
   * Identify wallet IDs based on the string and direction.
   *
   * @param string $match
   *   A fragment of the wallet name.
   *
   * @return Drupal\Core\Entity\Query\Sql\Query
   *   An entity query object, unexecuted
   */
  public function walletsQuery($match = NULL) {
    $settings = &$this->configuration['handler_settings'];

    $query = $this->entityTypeManager->getStorage('mcapi_wallet')->getQuery();
    if ($match) {
      $query->condition('name', '%' . $this->database->escapeLike($match) . '%', 'LIKE');
    }
    if (isset($settings['exclude'])) {
      $query->condition('wid', (array)$settings['exclude'], 'NOT IN');
    }
    // restriction is only set by the transaction fields otherwise not present.
    // Looks like currentuser isn't even used.
    if (isset($settings['restriction']) and $settings['restriction'] ==  static::RESTRICTION_MINE) {
      $query->condition('holder_entity_type', 'user');
      $query->condition('holder_entity_id', $this->currentUser->id());
    }
    return $query;
  }


  /**
   * This is only called in MassPay::buildEntity
   *
   * @param array $ids
   *
   * @return array
   *   All the wallets not passed
   */
  public function inverse(array $ids = []) {
    $entities = $this->getReferenceableEntities();
    return array_diff(array_keys($entities['mcapi_wallet']), $ids);
  }
}
